@@ -26,6 +26,7 @@
 #include "gcr-internal.h"
 #include "gcr-importer.h"
 #include "gcr-marshal.h"
+#include "gcr-oids.h"
 #include "gcr-parser.h"
 #include "gcr-types.h"
 
@@ -136,10 +137,12 @@ struct _GcrParserPrivate {
 	GTree *specific_formats;
 	gboolean normal_formats;
 	GPtrArray *passwords;
-	
+
 	GckAttributes *parsed_attrs;
 	const gchar *parsed_desc;
 	gchar *parsed_label;
+	gconstpointer parsing_block;
+	gsize parsing_n_block;
 };
 
 G_DEFINE_TYPE (GcrParser, gcr_parser, G_TYPE_OBJECT);
@@ -179,40 +182,18 @@ static GQuark PEM_PRIVATE_KEY;
 static GQuark PEM_PKCS7;
 static GQuark PEM_PKCS12;
 
-/* 
- * OIDS
- */
-
-static GQuark OID_PKIX1_RSA;
-static GQuark OID_PKIX1_DSA;
-static GQuark OID_PKCS7_DATA;
-static GQuark OID_PKCS7_SIGNED_DATA;
-static GQuark OID_PKCS7_ENCRYPTED_DATA;
-static GQuark OID_PKCS12_BAG_PKCS8_KEY;
-static GQuark OID_PKCS12_BAG_PKCS8_ENCRYPTED_KEY;
-static GQuark OID_PKCS12_BAG_CERTIFICATE;
-static GQuark OID_PKCS12_BAG_CRL;
-
 static void
 init_quarks (void)
 {
 	static volatile gsize quarks_inited = 0;
 
+	_gcr_oids_init ();
+
 	if (g_once_init_enter (&quarks_inited)) {
 
 		#define QUARK(name, value) \
 			name = g_quark_from_static_string(value)
-	 
-		QUARK (OID_PKIX1_RSA, "1.2.840.113549.1.1.1");
-		QUARK (OID_PKIX1_DSA, "1.2.840.10040.4.1");
-		QUARK (OID_PKCS7_DATA, "1.2.840.113549.1.7.1");
-		QUARK (OID_PKCS7_SIGNED_DATA, "1.2.840.113549.1.7.2");
-		QUARK (OID_PKCS7_ENCRYPTED_DATA, "1.2.840.113549.1.7.6");
-		QUARK (OID_PKCS12_BAG_PKCS8_KEY, "1.2.840.113549.1.12.10.1.1");
-		QUARK (OID_PKCS12_BAG_PKCS8_ENCRYPTED_KEY, "1.2.840.113549.1.12.10.1.2");
-		QUARK (OID_PKCS12_BAG_CERTIFICATE, "1.2.840.113549.1.12.10.1.3");
-		QUARK (OID_PKCS12_BAG_CRL, "1.2.840.113549.1.12.10.1.4");
-		
+
 		QUARK (PEM_CERTIFICATE, "CERTIFICATE");
 		QUARK (PEM_PRIVATE_KEY, "PRIVATE KEY");
 		QUARK (PEM_RSA_PRIVATE_KEY, "RSA PRIVATE KEY");
@@ -254,8 +235,15 @@ parsed_asn1_attribute (GcrParser *self, GNode *asn, const guchar *data, gsize n_
 }
 
 static void
-parsed_clear (GcrParser *self, CK_OBJECT_CLASS klass)
+parsing_begin (GcrParser *self,
+               CK_OBJECT_CLASS klass,
+               gconstpointer block,
+               gsize n_block)
 {
+	g_assert (GCR_IS_PARSER (self));
+	g_assert (block);
+	g_assert (n_block);
+
 	if (self->pv->parsed_attrs)
 		gck_attributes_unref (self->pv->parsed_attrs);
 	if (klass == CKO_PRIVATE_KEY)
@@ -263,10 +251,7 @@ parsed_clear (GcrParser *self, CK_OBJECT_CLASS klass)
 	else 
 		self->pv->parsed_attrs = gck_attributes_new ();
 	gck_attributes_add_ulong (self->pv->parsed_attrs, CKA_CLASS, klass);
-	
-	g_free (self->pv->parsed_label);
-	self->pv->parsed_label = NULL;
-	
+
 	switch (klass) {
 	case CKO_PRIVATE_KEY:
 		self->pv->parsed_desc = _("Private Key");
@@ -281,6 +266,9 @@ parsed_clear (GcrParser *self, CK_OBJECT_CLASS klass)
 		self->pv->parsed_desc = NULL;
 		break;
 	}
+
+	self->pv->parsing_block = block;
+	self->pv->parsing_n_block = n_block;
 }
 
 static void
@@ -298,6 +286,20 @@ parsed_attribute (GcrParser *self, CK_ATTRIBUTE_TYPE type, gconstpointer data, g
 	gck_attributes_add_data (self->pv->parsed_attrs, type, data, n_data);
 }
 
+static void
+parsing_end (GcrParser *self)
+{
+	g_assert (GCR_IS_PARSER (self));
+	self->pv->parsing_block = NULL;
+	self->pv->parsing_n_block = 0;
+	if (self->pv->parsed_attrs)
+		gck_attributes_unref (self->pv->parsed_attrs);
+	self->pv->parsed_attrs = NULL;
+	g_free (self->pv->parsed_label);
+	self->pv->parsed_label = NULL;
+	self->pv->parsed_desc = NULL;
+}
+
 static void
 parsed_ulong (GcrParser *self, CK_ATTRIBUTE_TYPE type, gulong value)
 {
@@ -385,7 +387,7 @@ parse_der_private_key_rsa (GcrParser *self, const guchar *data, gsize n_data)
 	if (!asn)
 		goto done;
 
-	parsed_clear (self, CKO_PRIVATE_KEY);
+	parsing_begin (self, CKO_PRIVATE_KEY, data, n_data);
 	parsed_ulong (self, CKA_KEY_TYPE, CKK_RSA);
 	parsed_boolean (self, CKA_PRIVATE, CK_TRUE);
 	res = GCR_ERROR_FAILURE;
@@ -410,12 +412,13 @@ parse_der_private_key_rsa (GcrParser *self, const guchar *data, gsize n_data)
 	
 	parsed_fire (self);
 	res = SUCCESS;
+	parsing_end (self);
 
 done:
 	egg_asn1x_destroy (asn);
 	if (res == GCR_ERROR_FAILURE)
 		g_message ("invalid RSA key");
-	
+
 	return res;
 }
 
@@ -433,7 +436,7 @@ parse_der_private_key_dsa (GcrParser *self, const guchar *data, gsize n_data)
 	if (!asn)
 		goto done;
 
-	parsed_clear (self, CKO_PRIVATE_KEY);
+	parsing_begin (self, CKO_PRIVATE_KEY, data, n_data);
 	parsed_ulong (self, CKA_KEY_TYPE, CKK_DSA);
 	parsed_boolean (self, CKA_PRIVATE, CK_TRUE);
 	ret = GCR_ERROR_FAILURE;
@@ -446,12 +449,13 @@ parse_der_private_key_dsa (GcrParser *self, const guchar *data, gsize n_data)
 		
 	parsed_fire (self);
 	ret = SUCCESS;
+	parsing_end (self);
 
 done:
 	egg_asn1x_destroy (asn);
 	if (ret == GCR_ERROR_FAILURE) 
 		g_message ("invalid DSA key");
-		
+
 	return ret;
 }
 
@@ -468,7 +472,6 @@ parse_der_private_key_dsa_parts (GcrParser *self, const guchar *keydata, gsize n
 	if (!asn_params || !asn_key)
 		goto done;
 
-	parsed_clear (self, CKO_PRIVATE_KEY);
 	parsed_ulong (self, CKA_KEY_TYPE, CKK_DSA);
 	parsed_boolean (self, CKA_PRIVATE, CK_TRUE);
 	ret = GCR_ERROR_FAILURE;
@@ -487,8 +490,8 @@ done:
 	egg_asn1x_destroy (asn_params);
 	if (ret == GCR_ERROR_FAILURE) 
 		g_message ("invalid DSA key");
-		
-	return ret;	
+
+	return ret;
 }
 
 /* -----------------------------------------------------------------------------
@@ -499,11 +502,11 @@ static gint
 parse_der_private_key (GcrParser *self, const guchar *data, gsize n_data)
 {
 	gint res;
-	
+
 	res = parse_der_private_key_rsa (self, data, n_data);
 	if (res == GCR_ERROR_UNRECOGNIZED)
 		res = parse_der_private_key_dsa (self, data, n_data);
-		
+
 	return res;
 }
 
@@ -533,12 +536,12 @@ parse_der_pkcs8_plain (GcrParser *self, const guchar *data, gsize n_data)
 	key_type = GCK_INVALID;
 
 	key_algo = egg_asn1x_get_oid_as_quark (egg_asn1x_node (asn, "privateKeyAlgorithm", "algorithm", NULL));
-  	if (!key_algo)
-  		goto done;
-  	else if (key_algo == OID_PKIX1_RSA)
-  		key_type = CKK_RSA;
-  	else if (key_algo == OID_PKIX1_DSA)
-  		key_type = CKK_DSA;
+	if (!key_algo)
+		goto done;
+	else if (key_algo == GCR_OID_PKIX1_RSA)
+		key_type = CKK_RSA;
+	else if (key_algo == GCR_OID_PKIX1_DSA)
+		key_type = CKK_DSA;
 
 	if (key_type == GCK_INVALID) {
   		ret = GCR_ERROR_UNRECOGNIZED;
@@ -554,7 +557,7 @@ parse_der_pkcs8_plain (GcrParser *self, const guchar *data, gsize n_data)
 	ret = SUCCESS;
 	
 done:
-	if (ret == SUCCESS) {		
+	if (ret == SUCCESS) {
 		switch (key_type) {
 		case CKK_RSA:
 			ret = parse_der_private_key_rsa (self, keydata, n_keydata);
@@ -562,18 +565,20 @@ done:
 		case CKK_DSA:
 			/* Try the normal sane format */
 			ret = parse_der_private_key_dsa (self, keydata, n_keydata);
-			
+
+			parsing_begin (self, CKO_PRIVATE_KEY, data, n_data);
 			/* Otherwise try the two part format that everyone seems to like */
 			if (ret == GCR_ERROR_UNRECOGNIZED && params && n_params)
 				ret = parse_der_private_key_dsa_parts (self, keydata, n_keydata, 
 				                                       params, n_params);
+			parsing_end (self);
 			break;
 		default:
 			g_message ("invalid or unsupported key type in PKCS#8 key");
 			ret = GCR_ERROR_UNRECOGNIZED;
 			break;
 		};
-		
+
 	} else if (ret == GCR_ERROR_FAILURE) {
 		g_message ("invalid PKCS#8 key");
 	}
@@ -612,7 +617,7 @@ parse_der_pkcs8_encrypted (GcrParser *self, const guchar *data, gsize n_data)
 
 	params = egg_asn1x_get_raw_element (egg_asn1x_node (asn, "encryptionAlgorithm", "parameters", NULL), &n_params);
 
-	parsed_clear (self, CKO_PRIVATE_KEY);
+	parsing_begin (self, CKO_PRIVATE_KEY, data, n_data);
 
 	/* Loop to try different passwords */                       
 	for (;;) {
@@ -658,7 +663,9 @@ parse_der_pkcs8_encrypted (GcrParser *self, const guchar *data, gsize n_data)
 		
 		/* We assume unrecognized data, is a bad encryption key */	
 	}
-		
+
+	parsing_end (self);
+
 done:
 	if (cih)
 		gcry_cipher_close (cih);
@@ -687,17 +694,20 @@ parse_der_pkcs8 (GcrParser *self, const guchar *data, gsize n_data)
 static gint
 parse_der_certificate (GcrParser *self, const guchar *data, gsize n_data)
 {
+	gchar *name = NULL;
 	GNode *asn;
-	gchar *name;
 
 	asn = egg_asn1x_create_and_decode (pkix_asn1_tab, "Certificate", data, n_data);
 	if (asn == NULL)
 		return GCR_ERROR_UNRECOGNIZED;
 
-	parsed_clear (self, CKO_CERTIFICATE);
+	parsing_begin (self, CKO_CERTIFICATE, data, n_data);
+
 	parsed_ulong (self, CKA_CERTIFICATE_TYPE, CKC_X_509);
 
-	name = egg_dn_read_part (egg_asn1x_node (asn, "tbsCertificate", "subject", "rdnSequence", NULL), "CN");
+	if (self->pv->parsed_label == NULL)
+		name = egg_dn_read_part (egg_asn1x_node (asn, "tbsCertificate", "subject", "rdnSequence", NULL), "CN");
+
 	egg_asn1x_destroy (asn);
 
 	if (name != NULL) {
@@ -707,7 +717,8 @@ parse_der_certificate (GcrParser *self, const guchar *data, gsize n_data)
 
 	parsed_attribute (self, CKA_VALUE, data, n_data);
 	parsed_fire (self);
-	
+
+	parsing_end (self);
 	return SUCCESS;
 }
 
@@ -783,7 +794,7 @@ parse_der_pkcs7 (GcrParser *self, const guchar *data, gsize n_data)
 	g_return_val_if_fail (oid, GCR_ERROR_FAILURE);
 
 	/* Outer most one must just be plain data */
-	if (oid != OID_PKCS7_SIGNED_DATA) {
+	if (oid != GCR_OID_PKCS7_SIGNED_DATA) {
 		g_message ("unsupported outer content type in pkcs7: %s", g_quark_to_string (oid));
 		goto done;
 	}
@@ -841,6 +852,41 @@ done:
 	return ret;
 }
 
+static gchar *
+parse_pkcs12_bag_friendly_name (GNode *asn)
+{
+	guint count, i;
+	GQuark oid;
+	GNode *node;
+	gconstpointer element;
+	gsize n_element;
+	GNode *asn_str;
+	gchar *result;
+
+	if (asn == NULL)
+		return NULL;
+
+	count = egg_asn1x_count (asn);
+	for (i = 1; i <= count; i++) {
+		oid = egg_asn1x_get_oid_as_quark (egg_asn1x_node (asn, i, "type", NULL));
+		if (oid == GCR_OID_PKCS9_ATTRIBUTE_FRIENDLY) {
+			node = egg_asn1x_node (asn, i, "values", 1, NULL);
+			if (node != NULL) {
+				element = egg_asn1x_get_raw_element (node, &n_element);
+				asn_str = egg_asn1x_create_and_decode (pkix_asn1_tab, "BMPString",
+				                                       element, n_element);
+				if (asn_str) {
+					result = egg_asn1x_get_bmpstring_as_utf8 (asn_str);
+					egg_asn1x_destroy (asn_str);
+					return result;
+				}
+			}
+		}
+	}
+
+	return NULL;
+}
+
 static gint
 handle_pkcs12_bag (GcrParser *self, const guchar *data, gsize n_data)
 {
@@ -849,7 +895,9 @@ handle_pkcs12_bag (GcrParser *self, const guchar *data, gsize n_data)
 	guint count = 0;
 	GQuark oid;
 	const guchar *element;
+	gchar *friendly;
 	gsize n_element;
+	guint i;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
 
@@ -865,41 +913,44 @@ handle_pkcs12_bag (GcrParser *self, const guchar *data, gsize n_data)
 	/* 
 	 * Now inside each bag are multiple elements. Who comes up 
 	 * with this stuff?
-	 * 
-	 * But this is where we draw the line. We only support one
-	 * element per bag, not multiple elements, not strange
-	 * nested bags, not fairy queens with magical wands in bags...
-	 * 
-	 * Just one element per bag.
 	 */
-	if (count >= 1) {
+	for (i = 1; i <= count; i++) {
 
-		oid = egg_asn1x_get_oid_as_quark (egg_asn1x_node (asn, 1, "bagId", NULL));
+		oid = egg_asn1x_get_oid_as_quark (egg_asn1x_node (asn, i, "bagId", NULL));
 		if (!oid)
 			goto done;
 
-		element = egg_asn1x_get_raw_element (egg_asn1x_node (asn, 1, "bagValue", NULL), &n_element);
+		element = egg_asn1x_get_raw_element (egg_asn1x_node (asn, i, "bagValue", NULL), &n_element);
 		if (!element)
 			goto done;
 
+		friendly = parse_pkcs12_bag_friendly_name (egg_asn1x_node (asn, i, "bagAttributes", NULL));
+		if (friendly != NULL)
+			parsed_label (self, friendly);
+
 		/* A normal unencrypted key */
-		if (oid == OID_PKCS12_BAG_PKCS8_KEY) {
+		if (oid == GCR_OID_PKCS12_BAG_PKCS8_KEY) {
 			r = parse_der_pkcs8_plain (self, element, n_element);
 			
 		/* A properly encrypted key */
-		} else if (oid == OID_PKCS12_BAG_PKCS8_ENCRYPTED_KEY) {
+		} else if (oid == GCR_OID_PKCS12_BAG_PKCS8_ENCRYPTED_KEY) {
 			r = parse_der_pkcs8_encrypted (self, element, n_element);
 			
 		/* A certificate */
-		} else if (oid == OID_PKCS12_BAG_CERTIFICATE) {
+		} else if (oid == GCR_OID_PKCS12_BAG_CERTIFICATE) {
 			r = handle_pkcs12_cert_bag (self, element, n_element);
 								
-		/* TODO: OID_PKCS12_BAG_CRL */
+		/* TODO: GCR_OID_PKCS12_BAG_CRL */
 		} else {
 			r = GCR_ERROR_UNRECOGNIZED;
 		}
-		 
-		if (r == GCR_ERROR_FAILURE || r == GCR_ERROR_CANCELLED) {
+
+		if (friendly != NULL)
+			parsed_label (self, NULL);
+
+		if (r == GCR_ERROR_FAILURE ||
+		    r == GCR_ERROR_CANCELLED ||
+		    r == GCR_ERROR_LOCKED) {
 			ret = r;
 			goto done;
 		}
@@ -944,8 +995,6 @@ handle_pkcs12_encrypted_bag (GcrParser *self, const guchar *data, gsize n_data)
 	if (!params)
 		goto done;
 
-	parsed_clear (self, 0);
-
 	/* Loop to try different passwords */
 	for (;;) {
 		
@@ -994,7 +1043,7 @@ handle_pkcs12_encrypted_bag (GcrParser *self, const guchar *data, gsize n_data)
 		
 		/* We assume unrecognized data is a bad encryption key */	
 	}
-		
+
 done:
 	if (cih)
 		gcry_cipher_close (cih);
@@ -1044,7 +1093,7 @@ handle_pkcs12_safe (GcrParser *self, const guchar *data, gsize n_data)
 		g_return_val_if_fail (bag, ret);
 
 		/* A non encrypted bag, just parse */
-		if (oid == OID_PKCS7_DATA) {
+		if (oid == GCR_OID_PKCS7_DATA) {
 
 			egg_asn1x_destroy (asn_content);
 			asn_content = egg_asn1x_create_and_decode (pkix_asn1_tab, "pkcs-7-Data", bag, n_bag);
@@ -1059,7 +1108,7 @@ handle_pkcs12_safe (GcrParser *self, const guchar *data, gsize n_data)
 			r = handle_pkcs12_bag (self, content, n_content);
 
 		/* Encrypted data first needs decryption */
-		} else if (oid == OID_PKCS7_ENCRYPTED_DATA) {
+		} else if (oid == GCR_OID_PKCS7_ENCRYPTED_DATA) {
 			r = handle_pkcs12_encrypted_bag (self, bag, n_bag);
 		
 		/* Hmmmm, not sure what this is */
@@ -1067,8 +1116,10 @@ handle_pkcs12_safe (GcrParser *self, const guchar *data, gsize n_data)
 			g_warning ("unrecognized type of safe content in pkcs12: %s", g_quark_to_string (oid));
 			r = GCR_ERROR_UNRECOGNIZED;
 		}
-		
-		if (r == GCR_ERROR_FAILURE || r == GCR_ERROR_CANCELLED) {
+
+		if (r == GCR_ERROR_FAILURE ||
+		    r == GCR_ERROR_CANCELLED ||
+		    r == GCR_ERROR_LOCKED) {
 			ret = r;
 			goto done;
 		}
@@ -1107,7 +1158,7 @@ parse_der_pkcs12 (GcrParser *self, const guchar *data, gsize n_data)
 		goto done;
 
 	/* Outer most one must just be plain data */
-	if (oid != OID_PKCS7_DATA) {
+	if (oid != GCR_OID_PKCS7_DATA) {
 		g_message ("unsupported safe content type in pkcs12: %s", g_quark_to_string (oid));
 		goto done;
 	}
@@ -1124,8 +1175,12 @@ parse_der_pkcs12 (GcrParser *self, const guchar *data, gsize n_data)
 	if (!content)
 		goto done;
 
+	parsing_begin (self, 0, data, n_data);
+
 	ret = handle_pkcs12_safe (self, content, n_content);
 
+	parsing_end (self);
+
 done:
 	g_free (content);
 	egg_asn1x_destroy (asn_content);
@@ -1223,25 +1278,25 @@ handle_encrypted_pem (GcrParser *self, GQuark type, gint subformat,
 		g_message ("missing encryption header");
 		return GCR_ERROR_FAILURE;
 	}
-	
-	/* Fill in information necessary for prompting */
-	parsed_clear (self, pem_type_to_class (type));
-	
+
+	res = GCR_ERROR_FAILURE;
 	for (;;) {
 
 		res = enum_next_password (self, &pstate, &password);
 		if (res != SUCCESS)
-			return res;
-		
+			break;
+
 		decrypted = NULL;
 		n_decrypted = 0;
 		
 		/* Decrypt, this will result in garble if invalid password */	
 		ret = egg_openssl_decrypt_block (val, password, -1, data, n_data, 
 		                                 &decrypted, &n_decrypted);
-		if (!ret)
-			return GCR_ERROR_FAILURE;
-			
+		if (!ret) {
+			res = GCR_ERROR_FAILURE;
+			break;
+		}
+
 		g_assert (decrypted);
 		
 		/* Unpad the DER data */
@@ -1255,10 +1310,10 @@ handle_encrypted_pem (GcrParser *self, GQuark type, gint subformat,
 
 		/* Unrecognized is a bad password */
 		if (res != GCR_ERROR_UNRECOGNIZED)
-			return res;		
+			break;
 	}
-	
-	return GCR_ERROR_FAILURE;
+
+	return res;
 }
 
 typedef struct {
@@ -1268,8 +1323,13 @@ typedef struct {
 } HandlePemArgs;
 
 static void
-handle_pem_data (GQuark type, const guchar *data, gsize n_data,
-                 GHashTable *headers, gpointer user_data)
+handle_pem_data (GQuark type,
+                 const guchar *data,
+                 gsize n_data,
+                 const gchar *outer,
+                 gsize n_outer,
+                 GHashTable *headers,
+                 gpointer user_data)
 {
 	HandlePemArgs *args = (HandlePemArgs*)user_data;
 	gint res = GCR_ERROR_FAILURE;
@@ -1279,21 +1339,26 @@ handle_pem_data (GQuark type, const guchar *data, gsize n_data,
 	/* Something already failed to parse */
 	if (args->result == GCR_ERROR_FAILURE)
 		return;
-	
+
+	/* Fill in information necessary for prompting */
+	parsing_begin (args->parser, pem_type_to_class (type), outer, n_outer);
+
 	/* See if it's encrypted PEM all openssl like*/
 	if (headers) {
 		val = g_hash_table_lookup (headers, "Proc-Type");
 		if (val && strcmp (val, "4,ENCRYPTED") == 0) 
 			encrypted = TRUE;
 	}
-	
+
 	if (encrypted)
 		res = handle_encrypted_pem (args->parser, type, args->subformat,  
 		                            headers, data, n_data); 
 	else
 		res = handle_plain_pem (args->parser, type, args->subformat, 
 		                        data, n_data);
-	
+
+	parsing_end (args->parser);
+
 	if (res != GCR_ERROR_UNRECOGNIZED) {
 		if (args->result == GCR_ERROR_UNRECOGNIZED)
 			args->result = res;
@@ -1636,8 +1701,8 @@ gcr_parser_class_init (GcrParserClass *klass)
 	                                G_TYPE_NONE, 0);
 	
 	init_quarks ();
-	_gcr_initialize ();
-	
+	_gcr_initialize_library ();
+
 	/* Check that the format tables are in order */
 	for (i = 1; i < G_N_ELEMENTS (parser_formats); ++i)
 		g_assert (parser_formats[i].format_id >= parser_formats[i - 1].format_id);
@@ -1692,26 +1757,28 @@ gcr_parser_parse_data (GcrParser *self, gconstpointer data,
                        gsize n_data, GError **error)
 {
 	ForeachArgs args = { self, data, n_data, GCR_ERROR_UNRECOGNIZED };
-	const gchar *message;
+	const gchar *message = NULL;
 	gint i;
 	
 	g_return_val_if_fail (GCR_IS_PARSER (self), FALSE);
 	g_return_val_if_fail (data || !n_data, FALSE);
 	g_return_val_if_fail (!error || !*error, FALSE);
 
-	/* Just the specific formats requested */
-	if (self->pv->specific_formats) { 
-		g_tree_foreach (self->pv->specific_formats, parser_format_foreach, &args);
-		
-	/* All the 'normal' formats */
-	} else if (self->pv->normal_formats) {
-		for (i = 0; i < G_N_ELEMENTS (parser_normal); ++i) {
-			if (parser_format_foreach ((gpointer)(parser_normal + i), 
-			                           (gpointer)(parser_normal + i), &args))
-				break;
+	if (data && n_data) {
+		/* Just the specific formats requested */
+		if (self->pv->specific_formats) {
+			g_tree_foreach (self->pv->specific_formats, parser_format_foreach, &args);
+
+		/* All the 'normal' formats */
+		} else if (self->pv->normal_formats) {
+			for (i = 0; i < G_N_ELEMENTS (parser_normal); ++i) {
+				if (parser_format_foreach ((gpointer)(parser_normal + i),
+							   (gpointer)(parser_normal + i), &args))
+					break;
+			}
 		}
 	}
-	
+
 	switch (args.result) {
 	case SUCCESS:
 		return TRUE;
@@ -1746,28 +1813,26 @@ gcr_parser_parse_data (GcrParser *self, gconstpointer data,
 void
 gcr_parser_format_enable (GcrParser *self, gint format_id)
 {
-	ParserFormat *format;
+	const ParserFormat *format;
+	guint i;
 
 	g_return_if_fail (GCR_IS_PARSER (self));
 
-	if (format_id == -1) {
-		if (self->pv->specific_formats)
-			g_tree_destroy (self->pv->specific_formats);
-		self->pv->specific_formats = NULL;
-		self->pv->normal_formats = TRUE;
-		return;
-	}
-
-	format = parser_format_lookup (format_id);
-	g_return_if_fail (format);
-
-	if (!self->pv->specific_formats) {
-		if (self->pv->normal_formats)
-			return;
+	if (!self->pv->specific_formats)
 		self->pv->specific_formats = g_tree_new (compare_pointers);
-	}
 
-	g_tree_insert (self->pv->specific_formats, format, format);
+	if (format_id != -1) {
+		format = parser_format_lookup (format_id);
+		g_return_if_fail (format);
+		g_tree_insert (self->pv->specific_formats,
+		               (gpointer)format, (gpointer)format);
+	} else {
+		for (i = 0; i < G_N_ELEMENTS (parser_formats); i++) {
+			format = &parser_formats[i];
+			g_tree_insert (self->pv->specific_formats, (gpointer)format,
+			               (gpointer)format);
+		}
+	}
 }
 
 /**
@@ -1868,6 +1933,27 @@ gcr_parser_get_parsed_label (GcrParser *self)
 	return self->pv->parsed_label;
 }
 
+/**
+ * gcr_parser_get_parsed_block:
+ * @self: a parser
+ * @n_block: a location to place the size of the block
+ *
+ * Get the raw data block that represents this parsed object. This is only
+ * valid during the GcrParser::parsed signal.
+ *
+ * Returns: The raw data block of the currently parsed item. The value is
+ *      owned by the parser and should not be freed.
+ */
+gconstpointer
+gcr_parser_get_parsed_block (GcrParser *self,
+                             gsize *n_block)
+{
+	g_return_val_if_fail (GCR_IS_PARSER (self), NULL);
+	g_return_val_if_fail (n_block, NULL);
+	*n_block = self->pv->parsing_n_block;
+	return self->pv->parsing_block;
+}
+
 /* ---------------------------------------------------------------------------------
  * STREAM PARSING
  */
